# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .common import Compiler
from ..tools import translate_flags


_powermake_flags_to_gcc_flags = {
    "-Weverything": ["-Wall", "-Wextra", "-fanalyzer"],
    "-ffuzzer": []
}

_powermake_flags_to_clang_flags = {
    "-fanalyzer": [],
    "-ffuzzer": ["-fsanitize=address,fuzzer"]
}

_powermake_flags_to_gnu_flags = {
    "-Weverything": ["-Wall", "-Wextra"],
    "-fanalyzer": [],
    "-ffuzzer": []
}


class CompilerGNU(Compiler):
    type = "gnu"
    obj_extension = ".o"
    translation_dict = _powermake_flags_to_gnu_flags

    def __init__(self, path: str = "cc"):
        super().__init__(path)

    @classmethod
    def format_args(self, defines: list, includedirs: list, flags: list = []):
        return [f"-D{define}" for define in defines] + [f"-I{includedir}" for includedir in includedirs] + translate_flags(flags, self.translation_dict)

    def basic_compile_command(self, outputfile: str, inputfile: str, args: list = []) -> list:
        return [self.path, "-c", "-o", outputfile, inputfile, *args]


class CompilerGNUPlusPLus(CompilerGNU):
    type = "gnu++"

    def __init__(self, path: str = "c++"):
        super().__init__(path)


class CompilerGCC(CompilerGNU):
    type = "gcc"
    translation_dict = _powermake_flags_to_gcc_flags

    def __init__(self, path: str = "gcc"):
        super().__init__(path)


class CompilerGPlusPlus(CompilerGNU):
    type = "g++"
    translation_dict = _powermake_flags_to_gcc_flags

    def __init__(self, path: str = "g++"):
        super().__init__(path)


class CompilerClang(CompilerGNU):
    type = "clang"
    translation_dict = _powermake_flags_to_clang_flags

    def __init__(self, path: str = "clang"):
        super().__init__(path)


class CompilerClangPlusPlus(CompilerGNU):
    type = "clang++"
    translation_dict = _powermake_flags_to_clang_flags

    def __init__(self, path: str = "clang++"):
        super().__init__(path)



# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .common import Archiver


class ArchiverMSVC(Archiver):
    type = "msvc"
    static_lib_extension = ".lib"

    def __init__(self, path: str = "lib"):
        super().__init__(path)

    def basic_archive_command(self, outputfile: str, inputfiles: set, args: list = []) -> list:
        return [self.path, "/nologo", *args, "/out:"+outputfile, *inputfiles]



import powermake


def on_build(config: powermake.Config):

    config.add_c_cpp_flags("-Weverything")

    config.add_shared_libs("pthread")

    files = powermake.get_files("*.c", "*.cpp")

    objects = powermake.compile_files(config, files)

    print(powermake.link_files(config, objects))


powermake.run("program_test", build_callback=on_build)



# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..tools import translate_flags
from .common import SharedLinker

_powermake_flags_to_gnu_flags = {
    "-ffuzzer": []
}

_powermake_flags_to_clang_flags = {
    "-ffuzzer": ["-fsanitize=address,fuzzer"]
}


class SharedLinkerGNU(SharedLinker):
    type = "gnu"
    shared_lib_extension = ".so"
    translation_dict = _powermake_flags_to_gnu_flags

    def __init__(self, path: str = "cc"):
        super().__init__(path)

    @classmethod
    def format_args(self, shared_libs: list, flags: list):
        return ["-l"+lib for lib in shared_libs] + translate_flags(flags, self.translation_dict)

    def basic_link_command(self, outputfile: str, objectfiles: set, archives: list = [], args: list = []) -> list:
        return [self.path, "-shared", "-o", outputfile, *objectfiles, *archives, *args]


class SharedLinkerGCC(SharedLinkerGNU):
    type = "gcc"

    def __init__(self, path: str = "gcc"):
        super().__init__(path)


class SharedLinkerGPlusPlus(SharedLinkerGNU):
    type = "g++"

    def __init__(self, path: str = "g++"):
        super().__init__(path)


class SharedLinkerClang(SharedLinkerGNU):
    type = "clang"
    translation_dict = _powermake_flags_to_clang_flags

    def __init__(self, path: str = "clang"):
        super().__init__(path)


class SharedLinkerClangPlusPlus(SharedLinkerGNU):
    type = "clang++"
    translation_dict = _powermake_flags_to_clang_flags

    def __init__(self, path: str = "clang++"):
        super().__init__(path)



# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import abc
import shutil


class Tool(abc.ABC):
    type = None

    def __init__(self, path):
        self._name = path
        self.reload()

    def is_available(self):
        return self.path is not None

    def reload(self):
        self.path = shutil.which(self._name)


def load_tool_tuple_from_file(conf: dict, tool_name: str, object_getter: callable, tool_list_getter: callable):
    if tool_name not in conf:
        return None

    if "path" in conf[tool_name] or "type" in conf[tool_name]:
        if "type" in conf[tool_name]:
            tool_type = conf[tool_name]["type"].lower()
        else:
            tool_type = "gnu"

        ObjectConstructor = object_getter(tool_type)
        if ObjectConstructor is None:
            raise ValueError("Unsupported %s type: %s\n\nShould be one of them: %s" % (tool_name, tool_type, " ".join(tool_list_getter())))

        if "path" in conf[tool_name]:
            tool_path = conf[tool_name]["path"]
        else:
            tool_path = None

        return (tool_path, ObjectConstructor)

    return None


def load_tool_from_tuple(tool_tuple, tool_name):
    tool: Tool = None
    if tool_tuple is not None:
        if tool_tuple[0] is None:
            tool = tool_tuple[1]()
        else:
            tool = tool_tuple[1](tool_tuple[0])

        if not tool.is_available():
            if tool_tuple[0] is None:
                tool_path = tool.type
            else:
                tool_path = tool_tuple[0]
            raise ValueError("The %s %s could not be found on your machine" % (tool_name, tool_path))
    return tool


def find_tool(object_getter: callable, *tool_types: str):
    for tool_type in tool_types:
        ObjectConstructor = object_getter(tool_type)
        if ObjectConstructor is None:
            continue
        tool: Tool = ObjectConstructor()
        if tool is not None and tool.is_available():
            return tool
    return None


def translate_flags(flags: list, translation_dict: dict):
    translated_flags = []
    for flag in flags:
        if flag in translation_dict:
            translated_flags.extend(translation_dict[flag])
        else:
            translated_flags.append(flag)

    return translated_flags



# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .common import Archiver


class ArchiverGNU(Archiver):
    type = "gnu"
    static_lib_extension = ".a"

    def __init__(self, path: str = "ar"):
        super().__init__(path)

    def basic_archive_command(self, outputfile: str, inputfiles: set, args: list = []) -> list:
        return [self.path, "-cr", outputfile, *inputfiles, *args]


class ArchiverAR(ArchiverGNU):
    type = "ar"

    def __init__(self, path: str = "ar"):
        super().__init__(path)


class ArchiverLLVM_AR(ArchiverGNU):
    type = "llvm-ar"

    def __init__(self, path: str = "llvm-ar"):
        super().__init__(path)



# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..tools import translate_flags
from .common import SharedLinker

_powermake_flags_to_msvc_flags = {
    "-m32": [],
    "-m64": [],
    "-ffuzzer": ["/fsanitize=address,fuzzer"]
}


class SharedLinkerMSVC(SharedLinker):
    type = "msvc"
    shared_lib_extension = ".dll"
    translation_dict = _powermake_flags_to_msvc_flags

    def __init__(self, path: str = "link"):
        super().__init__(path)

    @classmethod
    def format_args(self, shared_libs: list, flags: list):
        return [(lib if lib.endswith(".lib") else lib+".lib") for lib in shared_libs] + translate_flags(flags, self.translation_dict)

    def basic_link_command(self, outputfile: str, objectfiles: set, archives: list = [], args: list = []) -> list:
        return [self.path, "/DLL", "/nologo", *args, "/out:" + outputfile, *objectfiles, *archives]


class SharedLinkerClang_CL(SharedLinkerMSVC):
    type = "clang-cl"

    def __init__(self, path: str = "clang-cl"):
        super().__init__(path)



# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import abc

from ..tools import Tool


class Linker(Tool, abc.ABC):
    exe_extension = None

    def __init__(self, path):
        Tool.__init__(self, path)

    @classmethod
    @abc.abstractmethod
    def format_args(self, shared_libs: list, flags: list):
        return []

    @abc.abstractmethod
    def basic_link_command(self, outputfile: str, objectfiles: set, archives: list = [], args: list = []) -> list:
        return []
